#!/usr/local/python/3.3.2/bin/python3.3
#script-version: 5

import sys
import re
import os
import datetime
from CLI import CLI
import ipaddress
import pprint
import codecs
import binascii
import subprocess

INVALIDREQUIREDVALUE = -1
INVALIDVRF = -1
ERROR = -1
INVALIDMAC = 0
INVALIDIPSTR = '0.0.0.0'
INVALIDIPSTRWMASK = '0.0.0.0/24'
CLI_VERBOSE = False


class flowFwdValues():
    "Stores the tokens for a route entry"
    def __init__(self):
        self.fwdType = 0        
        self.ingressVlan = 0
        self.egressVlan = 0
        self.egressVNI = 0
        self.vrf = "N/A"
        self.vrfImportRT = "None"
        self.vrfImportl2RT = "None"
        self.dip = 0
        self.dipNetwork = ""
        self.ribPrefix = "N/A"
        self.bgpPrefix = "N/A"
        self.bgpLocalPeerIp = []
        self.bgpRemotePeerIp = []
        self.evpnPrefix = "N/A"
        self.evpnBestMacRouteFound = "N/A"
        self.evpnBestArpRouteFound = "N/A" 
        self.destLeafNodesIP = ""
        self.spineNodesIP = ""
        self.tunnelSIP = "None"
        self.tunnelDIP = "None"
        self.resolvedMacAddress = []
        self.egressInterface = "No Hardware entry"
        self.lpmHwFwdDecision = "N/A"
        self.nhHwFwdDecision = "N/A"
        self.l2exmHwFwdDecision = "N/A"
        self.alerts = 0
        

def extractFromCliOutput(cliString, matchString, tokenNum):
  global CLI_VERBOSE
  stringFoundInstances = 0
  requiredValueStr = INVALIDREQUIREDVALUE
  #if (CLI_VERBOSE):
    #print("CLI String: "+cliString)
  showCmd = CLI(cliString, False)
  showCmdLines = showCmd.get_output()
  for showCmdLine in showCmdLines:
    if matchString in showCmdLine:
      tokens = showCmdLine.split()
      requiredValueStr = tokens[tokenNum]
      stringFoundInstances = stringFoundInstances + 1
      #print ("Required Matches from CLI outout of "+cliString+" for a matchstring "+matchString+" is "+requiredValueStr+", String Found Instances =",stringFoundInstances)
  #print ("Required String Found = ",requiredValueStr)
  return (stringFoundInstances, requiredValueStr)


def extractMultipleItemsFromCmdOutput(cliString,matchString,matchToken,interestTokens):
# cliString - CLI, matchString - match specific string (* - ignore matching), 
# matchToken - token number of the matchString, if * - wildcard token for string match, if + or -, then represents an offset of match
# interestTokens - Extract the tokens mentioned in this list, "all" - match entire line

  global CLI_VERBOSE
  matchEntries = []
  matchedEntries = []
  matchFound = False
  #print("CLI_VERBOSE is ", CLI_VERBOSE)
  if (CLI_VERBOSE):
    print ("    DEBUG: CLI String: ", cliString)

  showCmd = CLI(cliString, False)
  showCmdLines = showCmd.get_output()
  for showCmdLine in showCmdLines:
    #print("TRACE:showCmdLine Line in current iteration: ", showCmdLine)
    cmdOutputLineTokens = []
    index = 0
    tokens = showCmdLine.split()
    try: #In case matchToken goes out of range for a line output
      #print ("Match String:", matchString, " and Token :", tokens[matchToken])
      if ((str(matchString) in showCmdLine) or (str(matchString) == "*")):
        if (str(matchToken) not in ["*", "+", "-"]) and (str(matchString) != "*"):
          if (str(matchString) == str(tokens[matchToken])):
            matchFound = True
          else:
            #Match not found on this line, move to the next for-loop iteration
            continue
        else:
          matchFound = True		
		  
        if (interestTokens == "all"):
          words = showCmdLine.split()
          for word in words:
            cmdOutputLineTokens.append(word)
          #print("cmdOutputLineTokens List for * = ", cmdOutputLineTokens)	                 
          
        else:
          if (str(matchToken) in ["+", "-"]):
            for token in tokens:
              if str(token) == str(matchString):
                index = tokens.index(token)
                #print("DEBUG: Match String Index is: ", index)
                for i, element in enumerate(interestTokens):
                  if (str(matchToken) == "+"):
                    interestTokens[i] = element + index
                  if (str(matchToken) == "-"):
                    interestTokens[i] = element - index
				
          for x in interestTokens:
            try:
              if (x < len(tokens)): #Boundary checks
                cmdOutputLineTokens.append(tokens[x])
                #print("cmdOutputLineTokens List = ", cmdOutputLineTokens)
                index+=1
              else:
                #print("DEBUG:Boundary checks violated for this line for interest token ",x,"...breaking out") 
                break
            except:
              #print ("Exception X is holding string: "+str(x)+ "Matched Instances = ", matchedInstances, "Wrapped Line consideration = ", wrappedLineConsideration)
              pass

        matchedEntries.append(cmdOutputLineTokens)
        #print ("Matched Entries List = ", matchedEntries)
      else:
        continue
        #print("Match Not Found in the command ", cliString, "for ", matchString)
    except IndexError:
      #print ("Match Token ", matchToken, "is out of bound...exiting the loop !!!")
      pass
  return (matchedEntries)



def inputParse(argv):
  da = ""
  sa = ""
  dip = ipaddress.ip_address(INVALIDIPSTR)
  dipNW = ipaddress.ip_address(INVALIDIPSTR)
  sip = ipaddress.ip_address(INVALIDIPSTR)
  sipNW = ipaddress.ip_address(INVALIDIPSTR)
  dstport = "N/A" #Current version does not support this argument
  srcport = "N/A" #Current version does not support this argument
  intfvlan = 0
  inputifname = "N/A" #Current version does not support this argument
  global CLI_VERBOSE
  DEBUG = False
  MAC_PATTERN = "^([0-9A-Fa-f]{4}[.]){2}([0-9A-Fa-f]{4})$"
  

  if (len(sys.argv) >= 2):
    if (sys.argv[1] == "-verbose"):
      #print("Verbose Enabled")
      CLI_VERBOSE = True	
      DEBUG = True
      index = 1
    else:
      #print("Verbose Disabled")
      CLI_VERBOSE = False
      index = 0
      DEBUG = False
  else:
    print ("\nUsage: execute-script ipfabric-ingressNode-debug.py [-verbose] <da> <sa> <dip> <sip> <intfvlan>")
    print ("Example: execute-script ipfabric-ingressNode-debug.py aaaa.bbbb.cccc dddd.eeee.ffff 10.1.1.81/24 10.1.1.82/24 10\n")
    return (False, DEBUG, da, sa, dip, dipNW, sip, sipNW, intfvlan, inputifname, dstport, srcport)

 
  if (DEBUG):  
    if (len(sys.argv) < 7):
      #print ("Usage ./ipfabric-ingressNode-debug.py <da> <sa> <dip> <sip> <intfvlan> [inputifname] [dstport] [srcport]") - For future version
      print ("\nUsage: execute-script ipfabric-ingressNode-debug.py [-verbose] <da> <sa> <dip> <sip> <intfvlan>")
      print ("Example: execute-script ipfabric-ingressNode-debug.py aaaa.bbbb.cccc dddd.eeee.ffff 10.1.1.81/24 10.1.1.82/24 10\n")
      return (False, DEBUG, da, sa, dip, dipNW, sip, sipNW, intfvlan, inputifname, dstport, srcport)
  else:
    if (len(sys.argv) < 6):
      #print ("Usage ./ipfabric-ingressNode-debug.py <da> <sa> <dip> <sip> <intfvlan> [inputifname] [dstport] [srcport]") - For future version
      print ("\nUsage: execute-script ipfabric-ingressNode-debug.py [-verbose] <da> <sa> <dip> <sip> <intfvlan>")
      print ("Example: execute-script ipfabric-ingressNode-debug.py aaaa.bbbb.cccc dddd.eeee.ffff 10.1.1.81/24 10.1.1.82/24 10\n")
      return (False, DEBUG, da, sa, dip, dipNW, sip, sipNW, intfvlan, inputifname, dstport, srcport)  

  #print ("da "+sys.argv[1]+" sa "+sys.argv[2]+" dip "+sys.argv[3]+" sip "+sys.argv[4]+" src intf vlan "+sys.argv[5]+" inputifname "+sys.argv[6]+" dstport "+sys.argv[7]+" srcport "+sys.argv[8])

  print("\nINPUT PARAMETERS:")
  print("====================\n")
   
  dastr = sys.argv[index+1]
  dastr = str(dastr).lower()
  if re.match(MAC_PATTERN, dastr):
    #print("DA MAC valid format ")
    pass
  else:
    print("***ALERT: DA MAC invalid format...exiting..")
    return (False, DEBUG, da, sa, dip, dipNW, sip, sipNW, intfvlan, inputifname, dstport, srcport)

  sastr = sys.argv[index+2]
  sastr = str(sastr).lower()
  if re.match(MAC_PATTERN, sastr):
    #print("SA MAC valid format ")
    pass
  else:
    print("***ALERT: SA MAC invalid format...exiting..")
    return (False, DEBUG, da, sa, dip, dipNW, sip, sipNW, intfvlan, inputifname, dstport, srcport)


  dipstr = sys.argv[index+3]
  dipString,sep,mask = str(dipstr).partition('/')
  if (mask):
    try:
      dipWithMask = ipaddress.ip_interface(dipstr)
      dipNW = dipWithMask.network
      dip = ipaddress.ip_address(dipString)
      #print("DIP with Mask, DIP Network, DIP: "+str(dipWithMask), str(dipNW), str(dip)) 
    except ValueError:
      print("***ALERT: Invalid DIP Address entered, please correct..exiting...")
      return (False, DEBUG, da, sa, dip, dipNW, sip, sipNW, intfvlan, inputifname, dstport, srcport)	  
  else:
    print("***ALERT: Invalid DIP Address format, please use mask...exiting..")
    return (False, DEBUG, da, sa, dip, dipNW, sip, sipNW, intfvlan, inputifname, dstport, srcport)	

  sipstr = sys.argv[index+4]
  sipString,sep,mask = str(sipstr).partition('/')
  if (mask):
    try:
      sipWithMask = ipaddress.ip_interface(sipstr)
      sipNW = sipWithMask.network
      sip = ipaddress.ip_address(sipString)
      #print("SIP with Mask, SIP Network, SIP: "+str(sipWithMask), str(sipNW), str(sip))
    except ValueError:
      print("***ALERT: Invalid SIP Address entered, please correct..exiting...")
      return (False, DEBUG, da, sa, dip, dipNW, sip, sipNW, intfvlan, inputifname, dstport, srcport)	
  else:
    print("***ALERT: Invalid SIP Address format, please use mask...exiting..")
    return (False, DEBUG, da, sa, dip, dipNW, sip, sipNW, intfvlan, inputifname, dstport, srcport)	
  

  if (sip == dip):
    print("***ALERT: SIP should not be the same as DIP, please correct the flow...exiting")
    return (False, DEBUG, da, sa, dip, dipNW, sip, sipNW, intfvlan, inputifname, dstport, srcport)
  
  intfstr = sys.argv[index+5]
  intfvlan = int(intfstr)  
  formatVlan = str(intfvlan).zfill(4)
  #print ("Formatted Vlan is ", formatVlan)

  if (len(sys.argv) > (index+6)):

    inputifname = sys.argv[index+6]
  
    dstportstr = sys.argv[index+7]
    dstport = int(dstportstr)

    srcportstr = sys.argv[index+8]
    srcport = int(srcportstr)
  
  print ("DA = "+str(dastr), ", SA = "+str(sastr), ", DIP = "+str(dip), ", SIP = "+str(sip), ", Ingress Vlan = ",intfvlan)
  
  return (True, DEBUG, dastr, sastr, dip, dipNW, sip, sipNW, intfvlan, inputifname, dstport, srcport)

  
def vlan2IntVlanMacIPVrf (intfvlan, sip, dip):

  #intvlanValid = True
  vlanVNIValid = False
  vlanVNI = 0
  vrfImportRT = []
  intvlanMacStr = []
	
  #A. For L3FWD traffic only - config checks
  vrfValid, vrf = extractFromCliOutput("show runn rbridge int ve "+str(intfvlan), "vrf forwarding", 2)
  intvlanMacStrValid1, mac1 = extractFromCliOutput("show int vlan "+str(intfvlan), "Current address is", 6)
  intvlanMacStr.append(mac1)
  print("    VE MAC Address = ", str(mac1))
  if (intvlanMacStrValid1): #VE interface configured
    intvlanAnycastIPValid, intvlanIPRaw = extractFromCliOutput("show runn rbridge int ve "+str(intfvlan), "ip anycast-address", 2)
    if (intvlanAnycastIPValid): #Anycast IP Address configured
      intIPAddressConfigured = True
      intvlanMacStrValid2, mac2 = extractFromCliOutput("show ip anycast-gateway vrf "+str(vrf), "Gateway mac", 2)
      if intvlanMacStrValid2:
        intvlanMacStr.append(mac2)
        print("    VE Interface config: Anycast IP configured: "+str(intvlanIPRaw)+" Anycast MAC configured: "+str(mac2))
      else:
        print("    ***ALERT: Anycast Gateway looking to be non-operational")
    else:
      intvlanVeIPValid, intvlanIPRaw = extractFromCliOutput("show runn rbridge int ve "+str(intfvlan), "ip address", 2)
      if (intvlanVeIPValid):
        print("    VE Interface config: Anycast IP NOT found, instead FOUND VE IP address")
        intIPAddressConfigured = True
      else: #NO IP address configured
        print("    ***ALERT: Config Error: VE Interface is NOT configured with IP Address...exiting...")
        return(False, 0,0,0,0,0,0)
  else: #VE not configured - INVALID
    print("    ***ALERT: Config Error: VE Interface for vlan"+str(intfvlan)+" is NOT configured for this Flow...exiting...")
    return(False, 0,0,0,0,0,0)

  #B. Deriving VRF, VRF VNI, VNI VE, VRF Import RT
  print("    Extracting vrf, vni, VE information.....<please wait>..")
  #vrfValid, vrf = extractFromCliOutput("show runn rbridge int ve "+str(intfvlan), "vrf forwarding", 2)
  vrfVNIValid, vrfVNI = extractFromCliOutput("show runn rbridge vrf "+str(vrf), "vni", 1)
  if (vrfVNIValid):
    vlanVNIValid, vlanVNI = extractFromCliOutput("sh vlan brief | until \"vni "+str(vrfVNI)+"\" | incl ACTIVE | last 1", "ACTIVE", 0)
    vrfImportRTEntries = extractMultipleItemsFromCmdOutput("show running-config rbridge-id vrf | beg "+str(vrf)+" | beg \"address-family ipv4 unicast\" | until \"!\" | incl import", "route-target", 0, [2])
    if (len(vrfImportRTEntries)):
      for entry in vrfImportRTEntries:
        #print("Param A OR First element of list = ", entry[0])
        if (vrfImportRT):
          vrfImportRT = vrfImportRT+","+str(entry[0])
          #vrfImportRT.append(str(entry[0]) #Cannot use this as need to save it as a string & not a list
        else:
          vrfImportRT = str(entry[0])
      print("    Appended list of VRF Import RT configured for this VRF: ", vrfImportRT)
    else:
      print("    ***LOG-MSG: No RT related configuration found !")
  else:
    print("    ***LOG-MSG: No VNI configuration found under this VRF", str(vrf), " proceeding further as L3-VNI is non-mandatory for non-symmetric routing..")
    vrfVNI = 0; vlanVNI = 0 # Dummy number
  #print ("\n    DEBUG: Valid var returned values: VE Mac address, IP address, vrf, L3-VNI, L3-VNI-VE: ", intvlanMacStrValid, intIPAddressConfigured, vrfValid, vrfVNIValid, vlanVNIValid)
  
  #C. Checks for the derived values		
  if ((intvlanMacStrValid1 or intvlanMacStrValid2) and intIPAddressConfigured and vrfValid):    
    intvlanIP = ipaddress.IPv4Interface(intvlanIPRaw)
    print("\nDERIVED PARAMETERS:")
    print("====================\n") 
    print ("Interface VE IP Address = "+str(intvlanIP.ip)+", VRF = "+vrf+", Int VE MAC addr (My DA list) = "+str(intvlanMacStr))
    print ("L3VNI = ", vrfVNI, ", L3VNI VE = ", vlanVNI, " VRF Import L3-RT configured for this VRF: ", vrfImportRT)
    return (True, intvlanMacStr, intvlanIP, vrf, vrfImportRT, vrfVNI, vlanVNI)

  else: #Incomplete Configuration for VRF, VNI, VNI VE
    print("    ***ALERT: Inconsistency found for VRF, VNI, VNI VE.... \n")
    #print("    DEBUG: Inconsistency among: VE Mac address, IP address, vrf, L3-VNI, L3-VNI-VE: ", intvlanMacStrValid1, intIPAddressConfigured, vrfValid, vrfVNIValid, vlanVNIValid)
    return(False, 0,0,0,0,0,0)  
  
 
  
def checkForFwdingType (sip, dip, intvlanIP, da, intvlanMac):
  checkPass = True
  l2fwd = True
  fwdType = "l2fwd"

  
  if (not (sip in intvlanIP.network)): #Technically, does not apply for L2 Fwd traffic, but good to flag such a case
    print ("\n***ALERT - sip "+str(sip)+" is not in the VE network, "+str(intvlanIP.network)+" may want to check...continuing..")
    checkPass = False
  else:
    if (dip in intvlanIP.network):
      l2fwd = True #L2 Switched packet
      if (da in intvlanMac):
        print ("\n***ALERT - l2fwd traffic should not be sent to myDA, please use Host DA - exiting!")
        checkPass = False
      else:
        print ("    Comment: <As dip "+str(dip)+" is in src network, "+str(intvlanIP.ip)+" thus forwarding type is L2FWD>")
    else:
      l2fwd = False #L3 Routed Packet
      fwdType = "l3fwd"
      if (da not in intvlanMac):
        print ("\n***ALERT - L3VPN traffic should be sent to myDA, please use correct da - exiting!")
        checkPass = False
      else:
        print ("    Comment: <As dip "+str(dip)+" is not in src network, "+str(intvlanIP.ip)+" thus forwarding type is L3VPN>")
  print ("    FORWARDING TYPE = ", fwdType)
  return (fwdType, checkPass, l2fwd)


def checkMacAddr (da, intfvlan):

  print("Checking MAC-table..if the MAC-table is long, this output will take more time.....")
  daEntries = extractMultipleItemsFromCmdOutput("show mac-address-table vlan "+str(intfvlan)+" | incl "+str(da), da, 1, [4,5])
  if (len(daEntries)):
    daFound = True
    print("    DA FOUND: Number of DA Entries in mac-table for DA ",str(da)," : ",len(daEntries))
    for entry in daEntries:
      ##intType = entry.ParamA; intNum = entry.ParamB
      intType = entry[0]; intNum = entry[1]
      print("    DA Entry Details : Interface Type:",str(intType), "Interface Number :", str(intNum)) 

  else:
    daFound = False; intType = 0; intNum = 0
    print("    MAC entry NOT FOUND in mac-table...")
   
  return (daFound, intType, intNum)


def bgpNeighborCheck (flow):

  print("\n\nIDENTIFYING THE EVPN BGP PEERS:")
  print("===================================\n")
  
  #STEP0: CHECK BGP PEERING & NOTE DOWN PEERS:

  print("Checking BGP table for Peering:")    
  bgpPeerEntries = extractMultipleItemsFromCmdOutput("show bgp evpn neighbors | incl \"Local host|Remote host\"", "*", 0, [0,1,2,3,4])  
  if (len(bgpPeerEntries)):
    print("    Total Number of BGP Peers:", int(len(bgpPeerEntries)/2))
    for peer in bgpPeerEntries:
      if (str(peer[0])+" "+str(peer[1]) == "Local host:"):
        #print("Local BGP Peering IP Address: ",peer[2].strip(','))
        flow.bgpLocalPeerIp.append(peer[2].strip(','))
      if (peer[0]+" "+peer[1] == "Remote host:"):
        #print("Remote BGP Peering IP Address: ",peer[2].strip(','))
        flow.bgpRemotePeerIp.append(peer[2].strip(','))

    for peers in flow.bgpRemotePeerIp[:]: 	
      index = flow.bgpRemotePeerIp.index(peers)
      #print("Peer for this iteration = ", peers, "Index = ", index)
      isEvpnPeerEntry = extractMultipleItemsFromCmdOutput("show ip bgp neighbors "+flow.bgpRemotePeerIp[index]+" | incl \"Peer Negotiated L2VPN EVPN address family\"", "*", 0, [0])
      #print("isEvpnPeerEntry = ",isEvpnPeerEntry)
      if (isEvpnPeerEntry):
        print("    eVPN Local BGP Peering IP: ", flow.bgpLocalPeerIp[index], "eVPN Remote BGP Peering IP: ", flow.bgpRemotePeerIp[index])
      else:
        #print("    Removing non-eVPN peering: Local BGP Peering IP: ", flow.bgpLocalPeerIp[index], "Remote BGP Peering IP: ", flow.bgpRemotePeerIp[index])
        flow.bgpRemotePeerIp.remove(flow.bgpRemotePeerIp[index]); flow.bgpLocalPeerIp.remove(flow.bgpLocalPeerIp[index])
  return(flow)


def macRouteCheckInEvpnBgpTable (flow, da):

  print("\n\nL2FWD / ASYMMETRIC RTG: CHECKING THE BGP MAC ROUTE EVPN TABLE:")
  print("===================================================================\n")
  flow.evpnBestMacRouteFound = False		
  #CHECK BGP eVPN TABLE FOR MAC ROUTE:
  i=0
  #print("\nChecking BGP EVPN table for the MAC Route existence:")
  ecmpBgpEvpnEntries = extractMultipleItemsFromCmdOutput("show bgp evpn l2route type mac brief | incl \"\["+str(flow.egressVlan)+"\]\["+str(da)+"\]\"", "*", 0, [0,1])
  if (len(ecmpBgpEvpnEntries)):
    #print("    Route FOUND in BGP table...")
    for route in ecmpBgpEvpnEntries:
      i+=1
      print("    BGP Route",i,": ", str(route[0]), str(route[1]))
      if (bool(re.match("(\*[s]*>)", str(route[0])))):
        print("    Best MAC Route FOUND in the BGP EVPN Table: ")
        print("    BGP Best Route",i,": ", str(route[0]), str(route[1]))
        flow.evpnBestMacRouteFound = True
      elif (route[0] == "*"):
        print("    Non-best Route: ", str(route[0]), str(route[1]))

    rtEntries = extractMultipleItemsFromCmdOutput("show bgp evpn l2route type mac detail | beg \"\["+str(flow.egressVlan)+"\]\["+str(da)+"\]\" | until RT | last 1", "*", 0, "all")
    if (len(rtEntries)):
      rtList = []; line = " "; flow.vrfImportl2RT = ""; rtVlanMatch = False
      for entry in rtEntries:
        line = line.join(entry)
        #print("    Entry = "+str(line))
        rtList = re.findall(r"RT \b[0-9,:,\.]+\b", str(line))
        print("    RT List raw format: ", rtList)
        for i, element in enumerate(rtList):
          rtList[i] = str(rtList[i]).lstrip('RT ')
          if (flow.vrfImportl2RT):
            flow.vrfImportl2RT = flow.vrfImportl2RT + "," + str(rtList[i])
          else:
            flow.vrfImportl2RT = str(rtList[i])
          #=========
          rtAs,sep,rtVlan = str(rtList[i]).partition(':')
          #print ("    DEBUG INFO: ", rtAs, sep, rtVlan, flow.egressVlan)
          if (int(rtVlan) == int(flow.egressVlan)):
            print("    Verified RT Vlan", int(rtVlan), " MATCHES Egress Vlan", flow.egressVlan)
            rtVlanMatch = True
          else:
            # Commented to avoid raising alarm.
            #print("    ***ALERT: RT Vlan", int(rtVlan), " does NOT match the Egress Vlan", flow.egressVlan)
            #flow.alerts += 1
            if not rtVlanMatch:
              rtVlanMatch = False
            #========
        print("    L2 RT List extracted along with Destn Leaf Node AS number  :", rtList)
        print("    L2 RT extracted matches Egress Vlan  :", rtVlanMatch)		
        
  else:
    print("    ***ALERT: Route NOT Found in the BGP EVPN Table...")
    flow.alerts += 1
  return (flow)
		

def arpRouteCheckInEvpnBgpTable (flow):

  MAC_PATTERN2 = "([0-9A-Fa-f]{4}[.]){2}([0-9A-Fa-f]{4})" #To match occurrence in middle of string
  print("\n\nASYMMETRIC RTG: CHECKING THE BGP ARP ROUTE EVPN TABLE:")
  print("=============================================================\n")
  flow.evpnBestArpRouteFound = False		
  #CHECK BGP eVPN TABLE FOR ARP ROUTE:
  i=0
  #print("\nChecking BGP EVPN table for the ARP Route existence:")
  ecmpBgpEvpnEntries = extractMultipleItemsFromCmdOutput("show bgp evpn l2route type arp brief | incl "+str(flow.dip)+"]", "*", 0, [0,1])
  if (len(ecmpBgpEvpnEntries)):
    #print("    Route FOUND in BGP EVPN table...")
    for route in ecmpBgpEvpnEntries:
      i+=1
      #print("    BGP Route",i,": ", str(route[0]), str(route[1]))
      if (bool(re.match("(\*[s]*>)", str(route[0])))):
        print("    Best ARP Route FOUND in the BGP EVPN Table: ")
        print("    BGP Best Route",i,": ", str(route[0]), str(route[1]))
        mac = re.search( MAC_PATTERN2, str(route[1]))
        #print("    Extracted MAC string: "+str(mac.group()))
        flow.evpnBestArpRouteFound = True
      elif ( "*" in route[0]):
        print("    Non-best Route: ", str(route[0]), str(route[1]))
        mac = re.search( MAC_PATTERN2, str(route[1]))
    rtEntries = extractMultipleItemsFromCmdOutput("show bgp evpn l2route type arp detail | beg "+str(mac.group())+" | until RT | last 1", "*", 0, "all")
    if (len(rtEntries)):
      rtList = []; line = " "; flow.vrfImportl2RT = ""; rtVlanMatch = False
      for entry in rtEntries:
        line = line.join(entry)
        #print("    Entry = "+str(line))
        rtList = re.findall(r"RT \b[0-9,:,\.]+\b", str(line))
        #print("    RT List raw format: ", rtList)
        for i, element in enumerate(rtList):
          rtList[i] = str(rtList[i]).lstrip('RT ')
          #if (flow.vrfImportl2RT):
            #flow.vrfImportl2RT = flow.vrfImportl2RT + "," + str(rtList[i])
          #else:
            #flow.vrfImportl2RT = str(rtList[i])
		  
          #========
          rtAs,sep,rtVlan = str(rtList[i]).partition(':')
          if (int(rtVlan) == int(flow.egressVlan)):
            print("    Verified RT Vlan", int(rtVlan), "MATCHES Egress Vlan",  flow.egressVlan)
            rtVlanMatch = True
            matchedRT = str(rtList[i])
            flow.vrfImportl2RT = str(rtList[i])
          else:
            if not rtVlanMatch:
              rtVlanMatch = False
            #Alert is commented to avoid raising it as an alarm. 
            #print("    ***ALERT: RT Vlan", int(rtVlan), "does NOT match the Egress Vlan", flow.egressVlan)
            #flow.alerts += 1
            #==========
        print("    L2 RT List extracted along with Destn Leaf Node AS number  :", rtList)
        print("    L2 RT extracted matches Egress Vlan  :", rtVlanMatch)		
        
  else:
    print("    ***ALERT: Route NOT Found in the BGP EVPN Table...")
    flow.alerts += 1
  return (flow)


  
def routeCheckInBgpVrfTable (flow):

  print("\n\nSYMMETRIC RTG: CHECKING THE BGP VRF TABLES:")
  print("===============================================\n")
  		
  #STEP1: CHECK BGP VRF TABLE:
  i=0
  print("\nChecking BGP VRF table for the Route existence:")
  ecmpBgpEntries = extractMultipleItemsFromCmdOutput("show ip bgp "+str(flow.dip)+" vrf "+str(flow.vrf), "*", 0, [0,1,2])
  if (len(ecmpBgpEntries)):
    #print("    Route FOUND in BGP table...")
    for route in ecmpBgpEntries:
      i+=1
      #print("    BGP Route",i,": ", str(route[0]), str(route[1]), str(route[2]))
      if (bool(re.match("(\*[s]*>)", str(route[0])))):
        #print("    Route FOUND in the BGP VRF Table: ")
        match = re.search( r"(\d{1,3}\.){3}\d{1,3}", str(route[1]))
        if (bool(match)):
          ip = match.group()
          print("    BGP Best Route: ", str(route[0]), str(ip), str(route[1]), str(route[2]))
          flow.bgpPrefix = ip
          flow.destLeafNodesIP = str(route[2]) #The Next Hop of the BGP VRF route would always point to Tunnel DIP as NH unchanged is configured.
        else:      
          flow.bgpPrefix = route[1]      
      elif (route[0] == "*"):
        print("    Non-best Route: ", str(route[0]), str(route[1]), str(route[2]))
        flow.bgpPrefix = route[2] #In case no best path, collect the route-prefix for the evpn check

  else:
    print("\n    ****ALERT: Route does NOT exist in BGP VRF table")
    flow.alerts += 1
	
  return (flow)

def routeCheckInBgpEvpnTable (flow):
		
  #STEP1: PRESENT IN BGP VRF TABLE, NOW CHECK BGP EVPN TABLE:
  i=0
  print("\nChecking BGP EVPN table for the Route Properties:")
  ecmpBgpEvpnEntries = extractMultipleItemsFromCmdOutput("show bgp evpn routes type ipv4-prefix "+str(flow.dipNetwork)+" tag 0 | incl NEXT_HOP", "NEXT_HOP:", 0, [1,5])
  if (len(ecmpBgpEvpnEntries)):
    #print("    Route FOUND in BGP EVPN table & the Number of Paths:", len(ecmpBgpEvpnEntries))
    print("    Route FOUND in BGP EVPN table..checking further.. ")
    #flow.evpnPrefix = flow.bgpPrefix
    flow.evpnPrefix = flow.dipNetwork
    for route in ecmpBgpEvpnEntries:
      i+=1
      #print("    BGP Route",i,":", "Advertising Spine Node:", str(route[1].strip(',')), " Advertising Egress Leaf Node:", str(route[0].strip(',')))
      if (str(route[1].strip(',')) != str(route[0].strip(','))):
        if (str(route[1].strip(',')) not in flow.spineNodesIP):
          flow.spineNodesIP = flow.spineNodesIP+" "+str(route[1].strip(','))
        #if (str(route[0].strip(',')) not in flow.destLeafNodesIP):
          #flow.destLeafNodesIP = flow.destLeafNodesIP+" "+str(route[0].strip(','))
          pass
      else:
        print("    ***ALERT: As the next hop & peer node is the same, the next-hop-unchanged is NOT configured on Spine Node")
        flow.alerts += 1
    #print("    BGP Route:", "Advertising Spine Node:", str(route[1].strip(',')))
    #print("    Advertising Egress Leaf Node:", str(route[0].strip(',')))

    #STEP2: CHECK THE RT FOR THE EVPN ROUTES & COMPARE WITH LOCAL
    i=0; doesMatchRT = True
    print("\nChecking BGP EVPN table for the RT values:")
    ecmpBgpEvpnEntries = extractMultipleItemsFromCmdOutput("show bgp evpn routes type ipv4-prefix "+str(flow.dipNetwork)+" tag 0 | incl RT", "RT", 2, [3])
    if (len(ecmpBgpEvpnEntries)):
      for route in ecmpBgpEvpnEntries:
        i+=1
        #print("    BGP Route",i,":", "RT:", str(route[0]))
        if (str(route[0]) in str(flow.vrfImportRT)):
          #print("    Matches the local RT, so should be imported into the VRF")
          pass
        else:
          print("    ****ALERT: Mismatch in RT, this ECMP route will NOT be imported into the VRF")
          flow.alerts += 1
          doesMatchRT = False
      if (doesMatchRT):
        print("    Route RT matches the local RT configured, thus should be imported into the VRF")

    else:
      print("    ***ALERT: RT NOT found in the the EVPN Route")
	
  else:
    print("    ***ALERT: BGP Route NOT found in eVPN table")
    flow.alerts += 1
  return (flow)
  
def printFlow (flow):

  print("\nQUICK GLANCE SUMMARY OF INGRESS NODE FLOW STRUCTURE:")
  print("=============================================================\n")
  print("Flow Forward Type                      : "+str(flow.fwdType))
  print("Flow Ingress Vlan                      : "+str(flow.ingressVlan))  
  print("Flow Egress Vlan                       : "+str(flow.egressVlan))
  print("Flow Egress Vlan VNI                   : "+str(flow.egressVNI))
  print("Flow VRF                               : "+str(flow.vrf))
  print("Flow VRF Import L3 RT                  : "+str(flow.vrfImportRT))
  print("Flow VRF Import L2 RT                  : "+str(flow.vrfImportl2RT))
  print("Flow DIP                               : "+str(flow.dip))  
  print("Flow IP Route found in RIB             : "+str(flow.ribPrefix))
  print("Flow IP Route found in BGP VRF Table   : "+str(flow.bgpPrefix))
  print("Flow IP Route found in BGP EVPN Table  : "+str(flow.evpnPrefix))
  print("Flow MAC Route found in BGP EVPN Table : "+str(flow.evpnBestMacRouteFound))
  print("Flow ARP Route found in BGP EVPN Table : "+str(flow.evpnBestArpRouteFound))
  print("Flow Egress LeafNodes IP's             : "+str(flow.destLeafNodesIP))
  print("Flow SpineNodes IP's                   : "+str(flow.bgpRemotePeerIp))
  print("Flow Egress Interface                  : "+str(flow.egressInterface)) 
  print("Flow Egress Int Tunnel SrcIP           : "+str(flow.tunnelSIP))
  print("Flow Egress Int Tunnel DestIP          : "+str(flow.tunnelDIP))  
  print("Flow Resolved MAC Address              : "+str(flow.resolvedMacAddress))    
  print("Flow L3 LPM Fwding Decision            : "+str(flow.lpmHwFwdDecision))    
  print("Flow L3 NH HW Fwding Decision          : "+str(flow.nhHwFwdDecision))
  print("Flow L2 HW Fwding Decision             : "+str(flow.l2exmHwFwdDecision))  
  print("Number of Alerts found                 : "+str(flow.alerts))
  print("=============================================================\n")
  print("Paste the below string on following Spine neighbor nodes for further tracing the packet path: ")
  print("  Reference Format: ./ipfabric-spineNode-debugv<x>.py <DIP> <DIP-prefix> <local-bgpPeerIP> <egress-vlan> \
                              \<fwd-type> <destn-Leaf-Node-IP's> <egressVNI> <MAC-Address> <tunnel-sip> <tunnel-dip> <RT>\n")
  
  if (("local" not in flow.fwdType) and flow.bgpRemotePeerIp):
    for x in range(len(flow.bgpRemotePeerIp)):
      for y in range(len(flow.resolvedMacAddress)):
        print("  Spine Node ("+str(flow.bgpRemotePeerIp[x])+"): ./ipfabric-spineNode-debugv<x>.py "+str(flow.dip)+" "+str(flow.dipNetwork)+" "+str(flow.bgpLocalPeerIp[x])+" "\
	    +" "+str(flow.egressVlan)+" "+str(flow.fwdType)+" "+str(flow.destLeafNodesIP)+" "+str(flow.egressVNI)+" "+str(flow.resolvedMacAddress[y])+" "\
        +str(flow.tunnelSIP)+" "+str(flow.tunnelDIP)+" "+str(flow.vrfImportRT)+"\n")
  return


  
def main():

  #Initialize the flow structure
  flow = flowFwdValues()
  
  #SECTION1: INPUT
  inputValid, DEBUG, da,sa, dip, dipNW, sip, sipNW, intfvlan, inputifname, dstport, srcport = inputParse(sys.argv)
  if (not inputValid):
    print("***ALERT: Input Error:...exiting..")
    return

  flow.dip = dip
  flow.dipNetwork = dipNW
  flow.ingressVlan = intfvlan
  #print ("Ingress Vlan = ", str(flow.ingressVlan), ", DIP = ", str(flow.dip))  
  
  #SECTION2: DERIVING VALUES & MAKING DECISIONS
  #SECTION2a: DERIVING VE IP, VRF, MAC, L3VNI & VE VNI FOR ROUTED TRAFFIC:
  
  # VLAN Validation
  intvlanValid, vlanFgl = extractFromCliOutput("show int vlan "+str(intfvlan), "Fine-Grained", 3)
  vlanValid, vlanActive = extractFromCliOutput("show vlan "+str(intfvlan), "VLAN"+str(intfvlan).zfill(4), 2)
  #print("Debug:", intvlanValid, vlanValid, vlanActive)
  if (not (intvlanValid and vlanValid)) and (vlanActive != "ACTIVE") :
    print("***ALERT: Config Error: VLAN"+str(intfvlan)+" not configured OR Active on the switch...exiting..")
    return
  else:
    print("\n    Verified the Vlan is valid & active..further checks..")
    ividValid, ivid = extractFromCliOutput("show system internal nsm gvlan "+str(intfvlan), str(intfvlan), 1)
    ivid_hex = str(hex(int(ivid))).zfill(4)
    ivid_hex = ivid_hex.lstrip("0x")
    #print("    Internal Vlan ID (IVID) = "+str(ivid), " and IVID in hex = ",ivid_hex)

  # Deriving Items
  intfvlanFound, intvlanMac, intvlanIP, vrf, vrfImportRT, vrfVNI, vlanVNI = vlan2IntVlanMacIPVrf (intfvlan, sip, dip)
  if (not intfvlanFound): 
    if (sipNW != dipNW): #L3-Fwding: Cannot proceed further with L3 flows processing
      print ("***ALERT - Invalid Configuration encountered... - exiting")
      return
    else: #L2 Fwding: Optional items for L2 flows, so will continue
      flow.fwdType = "l2fwd"
      print ("    ***ALERT - VE / VNI / RT not configured...thus validation will be limited..")
      flow.alerts += 1
      intvlanIP = 0
  else: 
    flow.vrf = vrf
    if (vrfImportRT):
      flow.vrfImportRT = vrfImportRT
    else:
      flow.vrfImportRT = "None"
	
 
  #SECTION2b: FURTHER VALIDITY CHECKING FOR THE FLOW BEING L2 SWITCHED OR L3 ROUTED
  # Assuming Ve is always configured for ingress vlan irrespective of fwding type
  
  if (intvlanIP):
    fwdType, IPvalidityCheckPass, l2fwd = checkForFwdingType (sip, dip, intvlanIP, da, intvlanMac)
    if (not IPvalidityCheckPass):
      print ("***ALERT - Failed Validity checks... - exiting")
      return
    flow.fwdType = fwdType

	
  #SECTION3: SUB-CLASSIFYING L2 FWDING TYPES
  if (flow.fwdType == "l2fwd"): #PACKET IS L2FWD
    print("\nMAC-TABLE CHECK:")
    print("====================\n")
    flow.egressVlan = intfvlan
    tunVlan = intfvlan
    #SECTION 3a: L2FWD: CHECK MAC TABLE FOR DA
    dafound, daEgressIntfType, daEgressIntf = checkMacAddr (da, intfvlan)
    if (not dafound):
      daflood = True
      flow.fwdType = "l2fwd-flood"
      print ("\n    L2 da "+da+" not found - packet will be flooded in vlan "+str(intfvlan))
    else:
      daflood = False
      print ("\n    DA "+da+" found & Egress Interface is "+str(daEgressIntfType)+" "+str(daEgressIntf))
      flow.egressInterface = str(daEgressIntfType)+" "+str(daEgressIntf)
      print("    Egress Interface = ", flow.egressInterface)      
      if (str(daEgressIntfType) == "Tu"):
        print ("    Packet will be switched over the Tunnel "+str(daEgressIntfType)+str(daEgressIntf))
        flow.fwdType = "l2fwd-Tu"
      else:
        flow.fwdType = "l2fwd-local"
        print ("    Packet will be switched locally on this Rbridge out of "+str(daEgressIntfType)+" "+str(daEgressIntf))
    #Check the eVPN Neighbors
    if ("local" not in flow.fwdType):
      flow = bgpNeighborCheck (flow)
      flow = macRouteCheckInEvpnBgpTable (flow,da)
   
  else: #PACKET IS L3 ROUTED
  
    #SECTION 3a: L3RTG: CHECK BGP TABLE
    flow = bgpNeighborCheck (flow)
 
    #SECTION 3b: L3RTG: DETERMINE IF SYMMETRIC OR ASSYMMETRIC RTG
    print("\nROUTE-TABLE CHECK:")
    print("====================\n")
    print("Checking VRF Routing Table .....")
    i = 0
    routeEcmpPaths = extractMultipleItemsFromCmdOutput("show ip route "+str(dip)+" vrf "+str(vrf)+" | beg Destination | excl Destination","*", 0, "all")
    #print("    ECMP Paths:", routeEcmpPaths)
    if (len(routeEcmpPaths)):
      print("    Route found for DIP: ", str(dip))
      print("    Number of ECMP Paths:", len(routeEcmpPaths))

 
      for route in routeEcmpPaths:
        i+=1
        #print("ECMP Route",i,": ", str(route[0]), str(route[1]), str(route[2]), str(route[3]), str(route[4]))
        try:    
          if (ipaddress.IPv4Network(route[0])):
            flow.ribPrefix = route[0]			
            #if (str(route[2])+str(route[3]) == "Ve"+vlanVNI):
            if ((route[1]) == "EVPN"):
              if (vlanVNI and vrfVNI):
                if (str(route[2])+str(route[3]) == "Ve"+vlanVNI):
                  print ("    SYMMETRIC ROUTING: It is an EVPN Route with Egress Port VE",str(route[3]),"being the same as VRF L3VNI VE",vlanVNI)
                  print("    ECMP Route",i,": ", str(route[0]), str(route[1]), str(route[2]), str(route[3]), str(route[5]))
                  flow.fwdType = "l3fwd-symmetric"
                  tunVlan = route[3]
                  flow.egressVlan = tunVlan
                else:
                  print("    LOG-MSG: Egress VE for this route is NOT matching L3-VNI, probably it is a VRF-Leak case...")
                  print ("    SYMMETRIC ROUTING: It is an EVPN Route with Egress Port VE",str(route[3]),"being NOT the same as VRF L3VNI VE",vlanVNI)
                  print("    ECMP Route",i,": ", str(route[0]), str(route[1]), str(route[2]), str(route[3]), str(route[5]))
                  flow.fwdType = "l3fwd-symmetric"
                  tunVlan = route[3]
                  flow.egressVlan = tunVlan
              else:
                print("    ***ALERT: VRF VNI & Vlan for the L3-VNI found to be NOT configured...exiting !")
                flow.alerts += 1
                return
                
            else:
              #print("    ASYMMETRIC / LOCAL ROUTING: Non-EVPN route with Egress Port VE",str(route[3])," NOT being VRF L3VNI VE",vlanVNI)
              tunVlan = route[3]
              flow.egressVlan = tunVlan           
              #if ((route.ParamF) == "D"):
              if ((route[1]) == "DIRECT"):
                print("    ASYMMETRIC ROUTING: Non-EVPN DIRECT route with Egress Port VE",str(route[3])," NOT being VRF L3VNI VE",vlanVNI)
                #print("    ASYMMETRIC ROUTING: Route Type = ",str(route[5]))
                print("    ECMP Route",i,": ", str(route[0]), str(route[1]), str(route[2]), str(route[3]), str(route[5]))
                flow.fwdType = "l3fwd-asymmetric"
                #Check if the MAC is local OR across the tunnel for the DIP VLAN
                #flow.fwdType = "l3fwd-asymmetric" OR flow.fwdType = "l3fwd-local-routing" OR flow.fwdType = "l3fwd-local-routing-flood"
                #show arp vrf <> (local routing) & show ip arp suppression-cache vlan <> (asymmetric - eVPN Rtg)
              else:
                print("    LOCAL PEERING: Non-EVPN Route Type = "+str(route[5])+", Next Hop = "+str(route[1]))
                print("    ECMP Route",i,": ", str(route[0]), str(route[1]), str(route[2]), str(route[3]), str(route[5]))
                routeType = str(route[5])

                flow.fwdType = "l3fwd-local-peering"
                print("\n    LOCAL PEERING: No check for ARP OR tunnels needed..")
        except:
          print("    ECMP Route",i,":              ", str(route[0]), str(route[1]), str(route[2]), str(route[4]))
          #pass    
    else:
      print("\n***ALERT: No Route Found for DIP: ", str(dip)," exiting.....")
      flow.alerts += 1
      flow.ribPrefix = "Not Found"
      flow.fwdType = "Drop"
      printFlow(flow)
      return

    if flow.fwdType in ("l3fwd-symmetric"):
      
      # CHECK SHOW IP ROUTE DETAIL
      print("\nOutput derived from show ip route detail: ")
      routePaths = extractMultipleItemsFromCmdOutput("show ip route detail vrf "+str(vrf)+" | begin "+str(flow.ribPrefix)+" | incl via","*via",0,[13,14,15])
      j = 0
      for route in routePaths:
        j+=1
        if (j <= len(routeEcmpPaths)):
          print("    Path ",j,": GW MAC Address: ",str(route[0]).strip(','), "Tunnel Interface: ", str(route[1])+str(route[2]).strip(')'))
          flow.resolvedMacAddress.append(route[0].strip(','))
          flow.egressInterface = str(route[1])+" "+str(route[2]).strip(')')
        else:
          break

      # CHECK BGP VRF TABLE
      flow = routeCheckInBgpVrfTable (flow)
      flow = routeCheckInBgpEvpnTable (flow)	  
		  
    #SECTION4: CHECKING FOR ARP ENTRIES     
    if flow.fwdType in ("l3fwd-asymmetric"):

      flow = routeCheckInBgpEvpnTable(flow)
      flow = macRouteCheckInEvpnBgpTable (flow, da)
      flow = arpRouteCheckInEvpnBgpTable (flow)

      print("\nCHECK ARP TABLE:")
      print("=====================\n")	  
      arpEntries = extractMultipleItemsFromCmdOutput("show ip arp suppression-cache vlan "+str(tunVlan)+" | incl "+str(dip), dip, 2, [3,4,5,7])
      if (len(arpEntries)):
        arpFound = True
        print("    ARP FOUND: Number of ARP Entries in suppression-cache for DIP",str(dip)," : ",len(arpEntries))
        for entry in arpEntries:
          flow.resolvedMacAddress.append(entry[0]); flow.egressInterface = str(entry[1])+" "+str(entry[2]) #; flow.resolvedMacAddress = entry[0]
          print("    ARP Entry Details : Mac Addr:",flow.resolvedMacAddress, "Interface :", str(flow.egressInterface), "Flags: ", str(entry[3])) 
          if (str(entry[1]) != "Tu"):
            flow.fwdType = "l3fwd-local-routing"
          else:
            pass
      else:
        arpFound = False
        print("    ARP entry NOT FOUND in suppression-cache...")
      
      arpEntries = extractMultipleItemsFromCmdOutput("show arp vrf "+str(vrf)+" | incl "+str(dip), dip, 0, [1,2,3,4,6])
      if (len(arpEntries)):
        arpFound = True
        print("    Number of Matching ARP Entries found:", len(arpEntries))
        print("    ARP Entry found for DIP: ", str(dip))    
        for entry in arpEntries:
          print("    ARP Entry Details : Mac Addr:",str(entry[0]), "Interface :", str(entry[1]), str(entry[2]), "MAC resolved: ", str(entry[3]), "Type :", str(entry[4]))
          macAddrARP = str(entry[0])          
          if (str(entry[1]) != "Tu"):
            flow.fwdType = "l3fwd-local-routing-arp" # Because ARP's are learnt over local interface & not tunnel
          else:
            pass
      else:
        if (arpFound != True):
          #arpFound = False
          print("    ARP entry NOT FOUND in Local VRF ARP table...") 

      if (arpFound):
        if (flow.fwdType == "l3fwd-asymmetric"):
          flow.fwdType = "l3fwd-asymmetric-arp"
        else:
          pass
      else:
          flow.fwdType = "l3fwd-asymmetric-no-arp"
          flow.resolvedMacAddress.append("None")

          
  #SECTION5: CHECKING THE TUNNEL STATES      
  if ("local" not in flow.fwdType):     
      print("\nTUNNEL STATE CHECK:")
      print("====================\n")
      #STEP A: CHECK TUNNEL VLAN VNI IS PROGRAMMED IN HW CORRECTLY

      ividValid, tunVlan_ivid = extractFromCliOutput("show system internal nsm gvlan "+str(tunVlan), str(tunVlan), 1)
      #print("    Tunnel Internal Vlan ID (IVID) = "+str(tunVlan_ivid))
	  
      num, hwVNI = extractFromCliOutput("unhide foscmd \nfibranne\n foscmd db tun show vni "+str(tunVlan_ivid)+" | incl VNI", "VNI", 3)
      #print("HW VNI = ", str(hwVNI), str(hwVNI).lstrip('='))
      swVNI = extractMultipleItemsFromCmdOutput("show vlan "+str(tunVlan)+" | incl vni", "vni", "+", [1])
      #print("SW VNI Entries =", swVNI)
      for entry in swVNI:
        if (entry):
          flow.egressVNI = str(entry[0])
      #print("SW VNI =", flow.egressVNI)

      if (str(hwVNI).lstrip('=') == flow.egressVNI):
        print("Verified the VNI programmed in Software for VLAN "+str(tunVlan)+" is same as programmed in Hardware")
      else:
        print("***ALERT: Verified the VLAN "+str(tunVlan)+" software VNI DOES NOT match with programmed in Hardware")
        flow.alerts += 1
        
      #STEP B: CHECK IF VXLAN TUNNELS ASSOCIATED WITH THE EGRESS VLAN IN SOFTWARE
      print ("\nTunnels associated in SOFTWARE for the vlan ", tunVlan, " on this Rbridge:")
      leafVlanTunnels = extractMultipleItemsFromCmdOutput("show vlan "+str(tunVlan)+" | incl Tu", "Tu", "*", "all")
      leafVlanTunnelsNumber = len(leafVlanTunnels)  
      if len(leafVlanTunnels):
        index = 0
        for entry in leafVlanTunnels:
          for z in range(0,len(entry)):
            if (str(entry[z]) == "Tu"):            			
              print ("    Tunnel ",index+1, " = Tu", str(entry[z+1]).strip('(t)'))
              index+=1
	  	  
	  
        #STEP B2: CHECK HARDWARE ENTRIES IF VXLAN TUNNELS ASSOCIATED WITH THE EGRESS VLAN
        print ("\nTunnels associated in HARDWARE for the vlan ", tunVlan, " on this Rbridge:")
        leafVlanHWTunnels = extractMultipleItemsFromCmdOutput("debug internal l2mgr debug mac dump vlan "+str(tunVlan_ivid), "tunnel", 0, "all")
        if (len(leafVlanHWTunnels)):
          print("    Found following tunnels associated with this vlan in HARDWARE")
          for entry in leafVlanHWTunnels:
            for x in range (2,len(entry)):
              a = str(entry[x][-4:])
              print("    Tunnel", int(a,16), end=" ")

          if ((len(entry)-2) == leafVlanTunnelsNumber):
            print("\n    Number of Tunnels programmed in HW are same as in Software")
          else:
            print("\n***ALERT: Number of Tunnels programmed in HW are NOT same as in Software")	
            flow.alerts += 1			

      else:     
        print ("    No Tunnels Found in the output")

          
      #STEP C: CHECK ALL TUNNELS & FLAG "DOWN" TUNNELS ON THIS RBRIDGE
      print ("\n\nTotal list of Tunnels that are Operationally UP on this Rbridge:")      
      tunnelEntries = extractMultipleItemsFromCmdOutput("show tunnel brief", "*", 0, [0,1,2,5])
      if (len(tunnelEntries)):
        #print("    Next Hop Entry FOUND in the Hardware NH Table")
        upTunnels = 0
        for entry in tunnelEntries:
          if (len(entry)):
            if (str(entry[0]) == "Tunnel"):
              #print("Tunnel Number: ", str(entry[1]).strip(','))
              tunnelNum = str(entry[1]).strip(',')
            elif (str(entry[0]) == "Source"):
              tunnelSrc = str(entry[2]).strip(',')
            elif (str(entry[0]) == "Admin"):
              #print("Oper State: ", str(entry[3]))
              tunnelState = str(entry[3])
            elif (str(entry[0]) == "Destination"):
              if (tunnelState.lower() == "up"):
                upTunnels+=1
                print("    Tunnel Number : "+tunnelNum+" Source IP: "+tunnelSrc+" Destination IP: "+str(entry[2])+" Oper State : "+tunnelState)
                #print("    DEBUG: Egress Int & Tunnel Numer: ", flow.egressInterfce, tunnelNum)
                if (flow.egressInterface == "Tu "+tunnelNum):
                  flow.tunnelSIP = tunnelSrc; flow.tunnelDIP = str(entry[2])
              else:
                print("\n***ALERT: Tunnel Tu"+tunnelNum+" found to be operationally down, please check reachability of DIP", str(entry[2]))
                flow.alerts += 1
            else:
              pass

       
      if not flow.destLeafNodesIP:
        flow.destLeafNodesIP = flow.tunnelDIP
      #STEP D: CHECK IF ALL UP TUNNELS ASSOCIATED WITH THIS VLAN
      if (leafVlanTunnelsNumber == upTunnels):
        print ("\nAll Operationally UP tunnels are associated with the vlan ",tunVlan)
      else:
        print ("\n***LOG-MSG: All Operationally UP tunnels are NOT associated with the vlan ",tunVlan)
        #flow.alerts += 1
      
  else:
    print("\nLOCAL PEERING2 OR LOCAL ROUTING: Tunnel checks bypassed..\n")
  
  
   
  #SECTION6: CHECKING IN HARDWARE

  #A. CHECK ALL RTR MAC's PROGRAMMED IN HW
    #bash-2.04# db 0/0 rte show exm ipv4addr used
  
  #B. CHECK FIB ARP OUTPUT
    #sw0# show arp slot 0
    #show arp slot 1     
      
  #C. CHECK NH LPM TABLE ENTRY
    #cz41-h06-m-r2:root> db 0/0 rte show lpm nhop all
    #sw0:L1/0: >db 7/0/1 rte show lpm v4 vrf 1
  
  #D. CHECK EXM IPV4 ADDRESS
    #bash-2.04# db 0/0 rte show exm ipv4addr used


  MAC_PATTERN = "^([0-9A-Fa-f]{4}[.]){2}([0-9A-Fa-f]{4})$"
	 
  if ("l3fwd-symmetric" in flow.fwdType):
    print("\nL3FWD SYMMETRIC: HARDWARE ROUTE-TABLE CHECK:")
    print("================================================\n")    
    
	#STEP1: CHECK FOR LPM ENTRIES
    print("\nCHECKING for LPM entry in HW....")
    prefix,sep,mask = str(flow.ribPrefix).partition('/')
    nhop_valid = False
    valid, isPresentInHardware = extractFromCliOutput("sh ip route slot 0 vrf "+str(flow.vrf)+" | incl "+str(flow.ribPrefix), str(flow.ribPrefix), 3)
    if (valid):
      print("    RIB Route programmed in Hardware : ", isPresentInHardware)
    else:
      print("    RIB Route NOT found")
    lpmEntries = extractMultipleItemsFromCmdOutput("show system internal fib 0 rib ip route vrf "+str(flow.vrf)+" "+str(flow.ribPrefix)+" | incl \"nhop_id \"", "chip", 0, [7,9,11])
    if (len(lpmEntries)):
      lpmEntriesFound = True
      print("    LPM Entry FOUND in Hardware for Prefix",str(flow.ribPrefix))
      for entry in lpmEntries:
        if ((int(entry[0]) == 0) and (mask == 32)):
          print("    This is a /32 masked prefix and not programmed in HW..possibly because no traffic..further nhop checks will not be needed")
        if ((int(entry[0]) == 1) and (entry[1].upper() == "TRAP")):
          print("    Subnet prefix programmed as TRAP in HW due to being advertised as default-gateway extended community..no further nhop checks")
        if (int(entry[0]) not in [0,1]):
          print("    LPM Entry Details : Nhop_Id: ",str(entry[0]), ", LPM HW Fwding Decision : ", str(entry[1]), ", Num of Paths: ", str(entry[2]))
          nhop_valid = True
        nhop_raw = str(entry[0])
        flow.lpmHwFwdDecision = str(entry[1])
		
		
      #STEP2: CHECK THE NHOP TABLE:
      if (nhop_valid):	  
        nhop_id = "0x"+str(hex(int(nhop_raw)))[2:].zfill(8)
        #print("    DEBUG: Nhop Index in hex: ", nhop_id )
        print("    The NHOP ID index obtained in hex: ",nhop_id)
	  
        print("\nCHECKING for NHOP entry in HW....")
        nhEntries = extractMultipleItemsFromCmdOutput("unhide foscmd \nfibranne\n foscmd db 0/0 rte show lpm nhop all | incl "+nhop_id, nhop_id, 0, "all")
        if (len(nhEntries)):
          nhEntriesFound = True
          macIndex = 0 #Index in the output
          print("    Next Hop Entry FOUND in the Hardware NH Table")
          for entry in nhEntries:
            tokens = entry
            for x in range (0,len(tokens)):
              if (re.match(MAC_PATTERN, tokens[x])):
                macIndex = x
            #print("    NH Entry Details : HW Index: ",str(entry[0]), ", Fwding Decision : ",str(entry[1]), ", MAC address of NH: ", str(entry[2]))
            print("    NH Entry Details : Internal Vlan (ivid): ",int(str(tokens[macIndex-3]), 16), ", Fwding Decision : ",str(tokens[macIndex-5]), ", MAC address of NH: ", str(tokens[macIndex]))  
            if (flow.nhHwFwdDecision == "N/A"):
              flow.nhHwFwdDecision = str(tokens[macIndex-5])
              #flow.resolvedMacAddress = str(tokens[macIndex]) Commenting because flow.resolvedMacAddress is populated by "show ip route detail"
              if (str(tokens[macIndex]) in flow.resolvedMacAddress):
                print("    HW Entry ", str(tokens[macIndex]), "  matched with Resolved MAC Address in SW")
              else:
                print("    ***ALERT: HW Entry ", str(tokens[macIndex]), "  DOES NOT match with Resolved MAC Address in SW")	
                flow.alerts += 1
            else:
              flow.nhHwFwdDecision = str(tokens[macIndex-5])
              #flow.resolvedMacAddress = flow.resolvedMacAddress+" "+str(tokens[macIndex]) Commenting because flow.resolvedMacAddress is populated by "show ip route detail"
              if (str(tokens[macIndex]) in flow.resolvedMacAddress):
                print("    HW Entry ", str(tokens[macIndex]), "  matched with Resolved MAC Address in SW")
              else:
                print("    ***ALERT: HW Entry ", str(tokens[macIndex]), "  DOES NOT match with Resolved MAC Address in SW")	
                flow.alerts += 1				

          #print("    DEBUG: List of Resolved MAC Addresses : ", flow.resolvedMacAddress)

          #STEP3: CHECK THE MAC TABLE FOR THE NHOP MAC		
          for mac in flow.resolvedMacAddress:
            print("\nCHECKING for MAC-Table entry for mac: "+str(mac)+" ....")
            macEntries = extractMultipleItemsFromCmdOutput("show mac-address-table vlan "+str(flow.egressVlan)+" | incl "+str(mac), flow.egressVlan, 0, [4,5])
            if (len(macEntries)):
              print("    Resolved NHOP Egress MAC "+str(mac)+" FOUND in the mac-address-table")
              for entry in macEntries:
                print("    Egress Interface is : "+str(entry[0])+str(entry[1]))
                flow.egressInterface = flow.egressInterface+"  "+str(entry[0])+str(entry[1])
            elif (flow.nhHwFwdDecision.upper() == "TRAP"):
              print("    Resolved NHOP NOT found in the mac-address-table, because NO Traffic for /32 host route")
            else:
              print("    ***ALERT: Resolved NHOP Egress MAC "+str(mac)+" NOT found in the mac-address-table")
              flow.alerts += 1
			
        else:
          nhEntriesFound = False
          flow.nhHwFwdDecision = "No NHOP Table Hardware Entry found"
          print("\n    ***ALERT: NO NH Entry found in Nhop table in Hardware / ASIC, ...")
          flow.alerts += 1
      else:
        nhEntriesFound = False
        flow.nhHwFwdDecision = "No NHOP Table Hardware Entry found"
        print("\n    ***ALERT: NO NH Entry found in Nhop table in Hardware / ASIC, ...")
        flow.alerts += 1
    else:
      lpmEntriesFound = False
      flow.lpmHwFwdDecision = "No LPM Table Hardware Entry found"
      print("\n    ***ALERT: NO Route-prefix LPM Entry found in Hardware / ASIC, ...")
      flow.alerts += 1
  
  elif ("l3fwd-asymmetric" in flow.fwdType):
    print("\nL3FWD ASYMMETRIC: HW MAC-ADDRESS TABLE CHECK:")
    print("==================================================\n")

    #STEP1: CHECK FOR ARP ENTRIES IN LOW-LEVEL SW
    print("\nCHECKING for ARP-Table in Low-level SW Slot0 entry for dip: "+str(flow.dip)+" ...."+flow.resolvedMacAddress[0])
    arpEntries = extractMultipleItemsFromCmdOutput("show arp slot 0 vrf "+str(flow.vrf)+" | incl \""+str(flow.dip)+"  \"", flow.resolvedMacAddress[0], 1, [5,6])
    if (len(arpEntries)):
      arpEntriesFound = True
      print("    ARP Entry FOUND in Low-level Software (slot0) for DIP:",str(flow.dip))
      for entry in arpEntries:
        print("    ARP resolved via : ",str(entry[0]), ", Programmed in Hardware : ", str(entry[1]).upper())

      #STEP2: CHECK FOR ARP ENTRIES IN ARP FIB			
      if (str(entry[1]).upper() == "YES"):
        print("\nCHECKING for FIB & L3-EXM Table for ARP entry for dip: "+str(flow.dip)+" ....")
        fibArpEntries = extractMultipleItemsFromCmdOutput("show system internal fib 0 arp ipv4 "+str(flow.dip)+" vrf "+str(flow.vrf)+" | incl MacResolved", "MacResolved", 0, [2,8,9,10])	  
        if (len(fibArpEntries)):
          print("    ARP Entry FOUND in FIB for DIP:",str(flow.dip))
          for entry in fibArpEntries:
            print("    MAC found in FIB: "+str(entry[0])+", Egress Rbridge : ",str(entry[1])+str(entry[2])+str(entry[3]))

          #Fishing data from the 2nd line of the output of the same CLI
          print("    CHECKING further using same command for the L3-EXM entry for ARP entry to obtain NHOP_ID ....")
          nhop_rawValid, nhop_raw = extractFromCliOutput("show system internal fib 0 arp ipv4 "+str(flow.dip)+" vrf "+str(flow.vrf)+" | incl \"Route Info\"", "", 5)	  
          nhop_id = str(nhop_raw).strip("NhopId[]")
          nhop_id = "0x"+str(hex(int(nhop_id)))[2:].zfill(8)
          print("    The NHOP ID index obtained in hex: ",nhop_id)

          #STEP3: CHECK THE NHOP TABLE FOR THE NHOP ID ENTRY
          print("\nCHECKING for NH Table for matching NHOP-ID entry ....")

          nhEntries = extractMultipleItemsFromCmdOutput("unhide foscmd \nfibranne\n foscmd db 0/0 rte show lpm nhop all | incl "+nhop_id, nhop_id, 0, "all")
          if (len(nhEntries)):
            nhEntriesFound = True
            macIndex = 0 #Index in the output
            print("    Next Hop Entry FOUND in the Hardware NH Table")
            for entry in nhEntries:
              tokens = entry
              for x in range (0,len(tokens)):
                #print("Token ",x, " = ", tokens[x])
                if (re.match(MAC_PATTERN, tokens[x])):
                  macIndex = x
                  #print("MAC Index found = ",x)
              #print("    NH Entry Details : HW Index: ",str(entry[0]), ", Fwding Decision : ",str(entry[1]), ", MAC address of NH: ", str(entry[2]))
              hwIvid = int(str(tokens[macIndex-3]), 16)
              print("    NH Entry Details : Internal Vlan (ivid): ",int(str(tokens[macIndex-3]), 16), ", Fwding Decision : ",str(tokens[macIndex-5]), ", MAC address of NH: ", str(tokens[macIndex]))
              if ((int(hwIvid) == int(tunVlan_ivid)) and (str(tokens[macIndex]) in flow.resolvedMacAddress)):
              #if (str(tokens[macIndex]) in flow.resolvedMacAddress):			  
                print("    Verified MAC progg & ivid in HW matches with SW")
              else:
                print("    ***ALERT: MAC HW Progg does NOT match with SW entries..", hwIvid, tunVlan_ivid, str(tokens[macIndex]), flow.resolvedMacAddress)
                flow.alerts += 1
              if (flow.nhHwFwdDecision == "N/A"):
                flow.nhHwFwdDecision = str(tokens[macIndex-5])
                #flow.resolvedMacAddress.append(str(tokens[macIndex])) #Commenting it as added during ARP table checks
              else:
                flow.nhHwFwdDecision = str(tokens[macIndex-5])
                #flow.resolvedMacAddress.append(str(tokens[macIndex]))	#Commenting it as added during ARP table checks			
                #print("    DEBUG: Resolved MAC Addresses : ", flow.resolvedMacAddress)
          else:
            print("\n    ***ALERT: NHOP-ID matching Entry NOT found in NHOP table ...")
            flow.alerts += 1			
        else:
          print("\n    ***ALERT: ARP Entry NOT found in FIB for DIP:",str(flow.dip))
          flow.alerts += 1
      else:
        print("\n    ***ALERT: ARP Entry in FIB indicating NOT programmed in Hardware (probably because no traffic) for DIP:",str(flow.dip))	
        flow.alerts += 1		
    else:
      print("\n    ***ALERT: ARP Entry NOT FOUND in Low-level Software (slot0) for DIP:",str(flow.dip))
      flow.alerts += 1
  elif ("l2fwd" in flow.fwdType):
    print("\nL2FWD: HW MAC-ADDRESS TABLE CHECK:")
    print("=====================================\n")
    print("If the MAC-table is long, this output will take more time.....")

    #STEP1: CHECK FOR L2EXM ENTRIES
    print("\nCHECKING for L2-EXM Table for MAC: "+str(da)+" ....")
    macEntries = extractMultipleItemsFromCmdOutput("unhide foscmd \nfibranne\n foscmd db 0/0 rte show exm l2 used | incl "+str(da), str(ivid_hex)+" "+str(da), "*", [0,1,2])
    if (len(macEntries)):
      for entry in macEntries:
        if (entry):
          macEntriesFound = True
          print("    MAC Entry FOUND in L2-EXM Hardware for Prefix",str(da)," and Vlan", flow.ingressVlan)
          print("    Hit-index = ", str(entry[0]))
          flow.l2exmHwFwdDecision = "Present"
          flow.resolvedMacAddress.append(str(da))	  
    else:
      print("\n    ***ALERT: MAC Entry NOT Found in L2-EXM Hardware for Prefix",str(da))
      flow.alerts += 1
      flow.l2exmHwFwdDecision = "Absent"
      flow.resolvedMacAddress.append(str(da))
 
 
  printFlow(flow) 
  return

        
  
  
             
if __name__ == "__main__":
  main()


